(**************************************************************)
(*   The Strong Completeness of Partial Monoidal BI           *)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import predicate.

Require Import words.
Require Import constraints.
Require Import pmo.

Require Import bi_simple.
 
Section Elem.

  Set Implicit Arguments.

  Variable L : Type.

  Inductive bi_basic (R : constraint L -> Prop) : constraint L -> Prop :=
    | In_bi_basic_1 : forall a b m, R (m ~~ m) -> a<> b -> ~ alphabet R a -> ~ alphabet R b -> bi_basic R (l a o l b ~~ m)
    | In_bi_basic_2 : forall a b m, R (m ~~ m) -> a<> b -> ~ alphabet R a -> ~ alphabet R b -> bi_basic R (l a o m ~~ l b)
    | In_bi_basic_3 : forall m b, R (m ~~ m) -> ~ alphabet R b -> bi_basic R (m ~~ l b)
    | In_bi_basic_4 : forall m, R (m ~~ m) -> bi_basic R (eps ~~ m).

  Proposition bi_basic_id R R' c : R ~ R' -> bi_basic R c -> bi_basic R' c.
  Proof.
    intros H1 [ ]; [ constructor 1 | constructor 2 | constructor 3 | constructor 4 ]; auto; try (apply H1; auto);
    match goal with [ H: ~ alphabet _ ?x |- ~ alphabet _ ?x ] =>  contradict H; generalize H; apply alphabet_inc, H1 end.
  Qed.
    
  Proposition bi_basic_eps R : is_pmo R -> bi_basic R (eps~~eps).
  Proof. constructor 4; apply ceq_eps with (1 := H). Qed.

  Definition bi_basic_pmo := basic_pmo bi_basic.
  Definition bi_simple_pmo := simple_pmo bi_basic.

  Hint Resolve bi_basic_id bi_basic_eps.

  Proposition bi_basic_inc_simple : bi_basic_pmo inc bi_simple_pmo.
  Proof. apply basic_inc_simple_pmo; auto. apply bi_basic_id. Qed.

  Proposition bi_simple_inc_is_pmo : bi_simple_pmo inc (@is_pmo _).
  Proof. apply simple_is_pmo. Qed.
    
End Elem.
