(**************************************************************)
(*   Copyright Dominique Larchey-Wendling [*]                 *)
(*                                                            *)
(*                             [*] Affiliation LORIA -- CNRS  *)
(**************************************************************)
(*      This file is distributed under the terms of the       *)
(*         CeCILL v2 FREE SOFTWARE LICENSE AGREEMENT          *)
(**************************************************************)

Require Import List.

Require Import dlist.

Set Implicit Arguments.

Reserved  Notation " c [ v ] "  (at level 40).

Section Term.

  Variable sig : nat -> Type.

  Unset Elimination Schemes.

  Inductive term : Type := in_term : forall n (c : sig n), dlist term n -> term.

  Notation " c [ v ] " := (in_term c v).

  Set Elimination Schemes.

  Section term_rect.

    Variable f : term -> Type.
    Variable Hf : forall n (c : sig n) (v : dlist _ n), dl_fall_t f v -> f (c[v]).

    Fixpoint term_rect t : f t :=
      match t as x return f x with
        | in_term n c v => 
          let fall := fix loop i (v : dlist _ i) : dl_fall_t f v :=
          match v in dlist _ i return dl_fall_t f v with
            | dl_nil        => in_dl_fall_t_0 f
            | dl_cons _ x v => in_dl_fall_t_1 x (term_rect x) (loop _ v)
          end in Hf c (fall _ v)
      end.

  End term_rect.

  Section term_rec.

    Variable f : term -> Set.
    Variable Hf : forall n (c : sig n) (v : dlist _ n), dl_fall_t f v -> f (c[v]).

    Definition term_rec t : f t.
    Proof. apply term_rect with (1 := Hf). Defined.

  End term_rec.

  Section term_ind.
  
    Variable f : term -> Prop.
    Variable Hf : forall n (c : sig n) (v : dlist _ n), dl_fall f v -> f (c[v]).

    Definition term_ind t : f t.
    Proof. 
      assert (forall n (c : sig n) (v : dlist _ n), dl_fall_t f v -> f (c[v])) as Hf'.
        intros; apply Hf, dl_fall_t_fall; auto.
      apply term_rect with (1 := Hf'). 
    Defined.

  End term_ind.

  Section term_recX.

    Variable X : Type.
    Variable f : forall n, sig n -> dlist X n -> X.

    Definition term_recX (t : term) : X.
    Proof. 
      induction t.
      apply dl_fall_t_dlist in X0.
      apply (f c X0).
    Defined.

    Proposition term_recX_prop n (c : sig n) v : term_recX (c[v]) = f c (dl_map term_recX v).
    Proof.
      simpl.
      f_equal. clear c.
      induction v; simpl; auto.
      rewrite IHv; auto.
    Qed.

  End term_recX.

  Variable f : forall n, sig n -> dlist Type n -> Type.
  
  Definition term_recS : term -> Type := term_recX f.

  Definition term_height := term_recX (fun n _ v => S (dl_max v)).

  Theorem term_height_prop n (c : sig n) v : term_height (c[v]) = S (dl_max (dl_map term_height v)).
  Proof. apply term_recX_prop. Qed.

End Term.

Notation " c [ v ] " := (in_term c v).

Section Term_subst.

  Variables sigX sigY : nat -> Type.

  Variable f : forall n (c : sigX n), dlist (term sigY) n -> term sigY.

  Definition term_subst := term_recX f.

  Theorem term_subst_prop n (c : sigX n) v : term_subst (c[v]) = f c (dl_map term_subst v).
  Proof.  apply term_recX_prop. Qed.

End Term_subst.

